const io = require("socket.io-client");

const Leap = require("leapjs");
const superagent = require("superagent");
const degrees = require("radians-degrees");
const URLS = require("./constants/urls.json");
const API_BASE_URL = URLS.API_BASE_URL;
const TOKEN_AUTH_URL = URLS.TOKEN_AUTH_URL;
const leapPlugins = require("leapjs-plugins");
const nj = require("numjs");

const dtw = require("@xesto/dtw");


const socket = io( API_BASE_URL );

const hackerModeGetAuthToken = ( apikey ) => {
  return superagent.post( TOKEN_AUTH_URL ).send({ apikey: apikey })
    .catch( err => { console.log("error authenticating") })
    .then(( response ) => {
      const token = response.body.token;
      return token;
    })
}

const getGestures = ( apikey ) => {
  return new Promise(( resolve, reject ) => {
    hackerModeGetAuthToken( apikey ).then( token => {
      socket.emit("GET_ALL_GESTURES", { jwt: token });
      socket.on("GESTURES_COMING_FROM_SERVER", ( response ) => {
        resolve( response );
      });
    })
  })
}

const classify = ( frames, gestures ) => {
  const whoWins = dtw.buildWhoWinsSorted( frames, gestures );
  const winner = whoWins.get(0, 1);
  const confidence = dtw.confidence( whoWins, winner,  5 );
  const name = winner.name;
  return { name: name, confidence: confidence }
}

const prepareFrame = ( frame ) => {
  const hand = frame.hands[0];
  const fingers = hand.fingers.map( finger => finger.dipPosition );
  const handData = [ hand.palmPosition, hand.pitch(), hand.roll(), hand.yaw() ];
  return [ ...handData, ...fingers ]
}
const palmSensitivityLimit = 1
const fingerSensitivityLimit = 1
const getAveragePalmPositionDifference = ( positions ) => {
  let distance;
  let sumOfDistances = 0;
  for (let i = 1; i < positions.length; i++) {
    let firstPosition = positions[i-1];
    let secondPosition = positions[i];
    let xPositionDifference = firstPosition[0] - secondPosition[0]
    let yPositionDifference = firstPosition[1] - secondPosition[1]
    let zPositionDifference = firstPosition[2]- secondPosition[2]
    distance = Math.sqrt(Math.pow(xPositionDifference,2) +
                        Math.pow(yPositionDifference,2) +
                        Math.pow(zPositionDifference,2) );
    sumOfDistances = sumOfDistances + distance;
  }
  return sumOfDistances/(positions.length - 1);
}

const getAverageFingerPositionDifference = ( positions) => {
  let distance;
  let sumOfDistances = 0;
  for (let i = 1; i < positions.length; i++) {
    let firstPosition = positions[i-1];
    let secondPosition = positions[i];
    let xPositionDifference = firstPosition.x - secondPosition.x;
    let yPositionDifference = firstPosition.y - secondPosition.y;
    let zPositionDifference = firstPosition.z - secondPosition.z;
    distance = Math.sqrt(Math.pow(xPositionDifference,2) +
                        Math.pow(yPositionDifference,2) +
                        Math.pow(zPositionDifference,2) );
    sumOfDistances = sumOfDistances + distance;
  }
  return sumOfDistances/(positions.length - 1);
}

function getPalmPositions(latestFramesOfExample) {
  return latestFramesOfExample.map(frame => frame.hands[0].palmPosition);
}

function isPalmMoving(palmPositions) {
  return getAveragePalmPositionDifference(palmPositions) > palmSensitivityLimit;
}

function groupFingers(latestExampleFrames) {
  const fingersData = latestExampleFrames.map(
    frame => frame.hands[0].fingers
  );
  const fingersCount = fingersData[0] ? fingersData[0].length : 0
  let eachFingerGrouped = [];
  for (let fingerNumber = 0; fingerNumber < fingersCount; fingerNumber++) {
    let dataCollectionOfFinger = []
    fingersData.forEach(fingers => {
      dataCollectionOfFinger = dataCollectionOfFinger.concat(fingers[fingerNumber].positions)
    })
    eachFingerGrouped = eachFingerGrouped.concat([dataCollectionOfFinger]);
  }
  return eachFingerGrouped
}

const areFingersMoving = ( groupedFingers ) => {
  return groupedFingers.some( finger => getAverageFingerPositionDifference( finger ) > fingerSensitivityLimit );
}

const isHandMoving = ( lastFewFrames ) => {
  if ( lastFewFrames.length === 0 ) {
    return false;
  }
  if( lastFewFrames.some( frame => frame.hands[0] === undefined )) {
    return false;
  }
  const palmPositions = getPalmPositions( lastFewFrames );
  const groupedFingers = groupFingers( lastFewFrames );
  return isPalmMoving( palmPositions ) || areFingersMoving( groupedFingers );
}

const buildController = ( classifier, options ) => {
  const controller = new Leap.Controller({ frameEventName: "animationFrame" });
  let frames = [];
  let isRecording = false;
  let motionDetectionFrames = []
  const logFrames = options.logFrames;
  
  Leap.plugin("xesto-wave", function ( options ) {
    const _super = this;
    return {
      frame: ( frame ) => {
        if( logFrames ) {
          console.log( frame );
        }
        if( motionDetectionFrames.length > 15 ) {
          motionDetectionFrames = motionDetectionFrames.slice(1, 15);
        }
        motionDetectionFrames.push( frame );
        const moving = isHandMoving( motionDetectionFrames );
        if ( moving ) {
          frames.push( prepareFrame( frame ));
        }
        if ( !moving && frames.length > 0 ){
            const classification = classifier( frames.reverse().slice(0, 180) );
            const gestureName = classififcation.name;
            const confidence = classification.confidence;
            
            console.log( gestureName, confidence );
            _super.emit( gestureName, confidence );
          }
        }
      }
    });
  controller.use("xesto-wave");
  return controller;
}

const getFingersDipArray = ( hand ) => {
  return _.map( hand.fingers, finger => finger.dipPosition );
}

const xesto = ( apikey ) => {
  return { 
   connect: ( options ) => {
      return getGestures( apikey ).then( response => {
        return response.gestures
      }).then( gestures => {
        return ( gesture ) => classify( gestures, gesture )
      }).then( classifier => {
        return buildController( classifier, options )
      }); 
    }
  }
}

module.exports = xesto;
